import tempfile
import os 

def creer_sommet( etiquette, fils=[] ):
    """
    Créer un sommet

    >>> s1 = creer_sommet( 'a' )
    >>> s1
    [None, 'a', []]
    >>> s2 = creer_sommet( 'b' )
    >>> s2
    [None, 'b', []]
    >>> s3 = creer_sommet( 'c' )
    >>> s3
    [None, 'c', []]
    >>> s4 = creer_sommet( 'd', [s1, s2, s3] )
    >>> s4
    [None, 'd', [[[...], 'a', []], [[...], 'b', []], [[...], 'c', []]]]
    >>> s5 = creer_sommet( 'e', [s4] )
    >>> s5
    [None, 'e', [[[...], 'd', [[[...], 'a', []], [[...], 'b', []], [[...], 'c', []]]]]]
    """
    res = [ None, etiquette, fils ]
    for f in fils:
        f[0] = res
    return res

def creer_arbre( racine ):
    """
    Créer un arbre.

    >>> s1 = creer_sommet( 'a' )
    >>> s2 = creer_sommet( 'b' )
    >>> s3 = creer_sommet( 'c' )
    >>> s4 = creer_sommet( 'd', [s1, s2, s3] )
    >>> s5 = creer_sommet( 'e', [s4] )
    >>> a = creer_arbre( s5 )
    >>> a
    [[None, 'e', [[[...], 'd', [[[...], 'a', []], [[...], 'b', []], [[...], 'c', []]]]]]]
    """
    return [racine]

def fils( arbre, sommet ):
    """
    Renvoie la liste des fils du sommet `sommet` de l'arbre `arbre`.

    >>> s1 = creer_sommet( 'a' )
    >>> s2 = creer_sommet( 'b' )
    >>> s3 = creer_sommet( 'c' )
    >>> s4 = creer_sommet( 'd', [s1, s2, s3] )
    >>> s5 = creer_sommet( 'e', [s4] )
    >>> a = creer_arbre( s5 )
    >>> fils( a, s4 ) == [s1, s2, s3]
    True
    """
    return sommet[2]

def pere( arbre, sommet ):
    """
    Renvoie le père du sommer `sommet` dans l'arbre `arbre`.

    >>> s1 = creer_sommet( 'a' )
    >>> s2 = creer_sommet( 'b' )
    >>> s3 = creer_sommet( 'c' )
    >>> s4 = creer_sommet( 'd', [s1, s2, s3] )
    >>> s5 = creer_sommet( 'e', [s4] )
    >>> a = creer_arbre( s5 )
    >>> pere( a, s4 ) is s5
    True
    >>> pere( a, s5 ) is None
    True
    """
    return sommet[0]

def racine( arbre ):
    """
    Renvoie la racine de l'abre.

    >>> s1 = creer_sommet( 'a' )
    >>> s2 = creer_sommet( 'b' )
    >>> s3 = creer_sommet( 'c' )
    >>> s4 = creer_sommet( 'd', [s1, s2, s3] )
    >>> s5 = creer_sommet( 'e', [s4] )
    >>> a = creer_arbre( s5 )
    >>> racine( a ) is s5
    True
    """
    return arbre[0]

def etiquette( arbre, sommet ):
    """
    Renvoie l'etiquette d'un sommet.

    >>> s1 = creer_sommet( 'a' )
    >>> s2 = creer_sommet( 'b' )
    >>> s3 = creer_sommet( 'c' )
    >>> s4 = creer_sommet( 'd', [s1, s2, s3] )
    >>> s5 = creer_sommet( 'e', [s4] )
    >>> a = creer_arbre( s5 )
    >>> etiquette( a, s2 )
    'b'
    """
    return sommet[1]


def taille_sous_arbre( arbre, sommet ):
    """
    Renvoie le nombre de sommet d'un sous arbre

    >>> s1 = creer_sommet( 'a' )
    >>> s2 = creer_sommet( 'b' )
    >>> s3 = creer_sommet( 'c' )
    >>> s4 = creer_sommet( 'd', [s1, s2, s3] )
    >>> s5 = creer_sommet( 'e', [s4] )
    >>> a = creer_arbre( s5 )
    >>> taille_sous_arbre( a, s4 )
    4
    """
    res = 1
    for f in fils( arbre, sommet ):
        res += taille_sous_arbre( arbre, f )
    return res

def taille_arbre( arbre ):
    """
    Renvoie le nombre de sommets de l'arbre.

    >>> s1 = creer_sommet( 'a' )
    >>> s2 = creer_sommet( 'b' )
    >>> s3 = creer_sommet( 'c' )
    >>> s4 = creer_sommet( 'd', [s1, s2, s3] )
    >>> s5 = creer_sommet( 'e', [s4] )
    >>> a = creer_arbre( s5 )
    >>> taille_arbre( a )
    5
    >>> r = creer_sommet( 'r' )
    >>> a = creer_arbre( r )
    >>> taille_arbre( a )
    1
    """
    return taille_sous_arbre( arbre, racine(arbre) )

def parcours_sous_arbre( arbre, sommet, p ):
    p.append( sommet )
    for f in fils( arbre, sommet ):
        parcours_sous_arbre( arbre, f, p )

def parcours_arbre( arbre, p ):
    """
    >>> s1 = creer_sommet( 'a' )
    >>> s2 = creer_sommet( 'b' )
    >>> s3 = creer_sommet( 'c' )
    >>> s4 = creer_sommet( 'd', [s1, s2, s3] )
    >>> s6 = creer_sommet( 'e' )
    >>> s5 = creer_sommet( 'f', [s4, s6] )
    >>> a = creer_arbre( s5 )
    >>> l = []
    >>> parcours_arbre( a, l )
    >>> res = [s5, s4, s1, s2, s3, s6]
    >>> len(l) == len( res )
    True
    >>> all( [ l[i] is res[i] for i in range(len(res)) ] )
    True
    """
    parcours_sous_arbre( arbre, racine(arbre), p )


def est_feuille( arbre, s ):
    return len( fils( arbre, s) ) == 0

def parcours_feuilles_sous_arbre( arbre, sommet, p ):
    """
    """
    if est_feuille( arbre, sommet ):
        p.append( sommet )
    else:
        for f in fils( arbre, sommet ):
            parcours_feuilles_sous_arbre( arbre, f, p )

def parcours_feuilles( arbre, p ):
    """
    Parcours l’arbre A et qui remplie la liste p avec les sommets de l’arbre 
    qui sont des feuilles.

    >>> s8 = creer_sommet( '8' )
    >>> s9 = creer_sommet( '9' )
    >>> s4 = creer_sommet( '4' )
    >>> s5 = creer_sommet( '5', [s8] )
    >>> s6 = creer_sommet( '6' )
    >>> s7 = creer_sommet( '7', [s9] )
    >>> s2 = creer_sommet( '2', [s4, s5] )
    >>> s3 = creer_sommet( '3', [s6, s7] )
    >>> s1 = creer_sommet( '1', [s2, s3] )
    >>> a = creer_arbre( s1 )
    >>> l = []; parcours_feuilles( a, l )
    >>> l == [s4, s8, s6, s9]
    True
    """
    parcours_feuilles_sous_arbre( arbre, racine(arbre), p )


def parcours_sommets_internes_sous_arbre( arbre, sommet, p ):
    """
    """
    if not est_feuille( arbre, sommet ):
        p.append( sommet )
        for f in fils( arbre, sommet ):
            parcours_sommets_internes_sous_arbre( arbre, f, p )

def parcours_sommets_internes( arbre, p ):
    """
    Parcours l’arbre A et qui remplie la liste p avec les sommets de l’arbre 
    qui sont des sommets internes.

    >>> s8 = creer_sommet( '8' )
    >>> s9 = creer_sommet( '9' )
    >>> s4 = creer_sommet( '4' )
    >>> s5 = creer_sommet( '5', [s8] )
    >>> s6 = creer_sommet( '6' )
    >>> s7 = creer_sommet( '7', [s9] )
    >>> s2 = creer_sommet( '2', [s4, s5] )
    >>> s3 = creer_sommet( '3', [s6, s7] )
    >>> s1 = creer_sommet( '1', [s2, s3] )
    >>> a = creer_arbre( s1 )
    >>> l = []; parcours_sommets_internes( a, l )
    >>> l == [s1, s2, s5, s3, s7]
    True
    """
    parcours_sommets_internes_sous_arbre( arbre, racine(arbre), p )

def arbre_to_dot( arbre ):
    res = ""
    sommets = []
    parcours_arbre( arbre, sommets )
    res += "\ndigraph G {"
    res += "\n\t"+'graph [ordering="out"];'
    for i in range( len(sommets) ) :
        s = sommets[i]
        res += "\n\tn" + str(i) + ' [ label="' + str(etiquette(arbre, s))+ '" ];'
    for i in range( len(sommets) ) :
        s = sommets[i]
        for f in fils( arbre, s ):
            for j in range( len(sommets) ) :
                if f is sommets[j]:
                    break
            res += "\n\tn" + str(i) + " -> " + "n" + str(j) + ";"
    res += "\n}"
    return res

def afficher_arbre( arbre ):
    """

    >>> s8 = creer_sommet( '8' )
    >>> s9 = creer_sommet( '9' )
    >>> s4 = creer_sommet( '4' )
    >>> s5 = creer_sommet( '5', [s8] )
    >>> s6 = creer_sommet( '6' )
    >>> s7 = creer_sommet( '7', [s9] )
    >>> s2 = creer_sommet( '2', [s4, s5] )
    >>> s3 = creer_sommet( '3', [s6, s7] )
    >>> s1 = creer_sommet( '1', [s2, s3] )
    >>> a = creer_arbre( s1 )
    >>> #afficher_arbre( a )
    """
    f = tempfile.NamedTemporaryFile( delete=False )
    f.write( bytes( arbre_to_dot( arbre ), 'UTF-8' ) )
    f.close()
    os.system( 'dotty ' + str(f.name) )
    os.remove( f.name )

if __name__ == "__main__":
    import doctest
    doctest.testmod() 
