import tempfile
import os 

def creer_sommet( etiquette, fils_gauche = None, fils_droit = None ):
    """
    Créer un sommet de l'arbre

    Entrée :
        fils_gauche : le fils gauche de l'arbre
        fils_droit : le ils droit de l'arbre
        etiquette : l'étiquettes associée au sommet

    Exemples :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
    """
    res = [ None, etiquette, [fils_gauche, fils_droit] ]
    if not fils_gauche is None:
        fils_gauche[0] = res
    if not fils_droit is None:
        fils_droit[0] = res
    return res

def creer_arbre_binaire( racine ):
    """
    Créer un arbre binaire.

    Entrée :
        racine : la racine de l'arbre.

    Exemples :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> ab = creer_arbre_binaire( s4 )
        >>> ab
        [[None, 4, [[[...], 8, [None, None]], [[...], 9, [None, None]]]]]
    """
    return [racine]

def taille_sous_arbre_binaire( arbre, sommet ):
    """
    Renvoie la taille d'un sous-arbre

    Entree :
        arbre - un arbre
        sommet - le sommet du sous-arbre

    Exemple :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> taille_sous_arbre_binaire( ab, s2 )
        6
    """
    if sommet is None:
        return 0
    return 1 + (
        taille_sous_arbre_binaire( arbre, fils_gauche(arbre, sommet) ) +
        taille_sous_arbre_binaire( arbre, fils_droit(arbre, sommet) )
    )

def taille_arbre_binaire( arbre ):
    """
    Renvoie la taille d'un arbre

    Entree :
        arbre - arbre

    Exemple :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> taille_arbre_binaire( ab )
        11
    """
    return taille_sous_arbre_binaire( arbre, racine(arbre) )

def fils_gauche( arbre, sommet ):
    """
    Renvoie le fils gauche d'un sommet dans un arbre

    Entrée : 
        arbre : l'arbre
        sommet : Un sommet

    Exemples :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> ab = creer_arbre_binaire( s4 )
        >>> fils_gauche( ab, s4 ) == s8
        True
        >>> fils_gauche( ab, s8 ) is None
        True
    """
    return sommet[2][0]

def fils_droit( arbre, sommet ):
    """
    Renvoie le fils droit d'un sommet dans un arbre

    Entrée : 
        arbre : l'arbre
        sommet : Un sommet

    Exemples :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> ab = creer_arbre_binaire( s4 )
        >>> fils_droit( ab, s4 ) == s9
        True
        >>> fils_droit( ab, s8 ) is None
        True
    """
    return sommet[2][1]

def fils( arbre, sommet ):
    """
    Renvoie la liste des fils d'un sommet

    Entrée :
        arbre - un arbre
        sommet - un sommet

    Exemple:
        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> fils( ab, s4 ) == [s8, s9]
        True
        >>> fils( ab, s12 ) == []
        True
    """
    res = []
    fg = fils_gauche( arbre, sommet )
    fd = fils_droit( arbre, sommet )
    if not fg is None:
        res.append(fg)
    if not fd is None:
        res.append(fd)
    return res

def pere( arbre, sommet ):
    """
    Renvoie le père d'un sommet

    Entrée : 
        arbre : l'arbre
        sommet : Un sommet

    Exemples :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> ab = creer_arbre_binaire( s4 )
        >>> pere( ab, s4 ) is None
        True
        >>> pere( ab, s8 ) is s4
        True
    """
    return sommet[0]

def racine( arbre ):
    """
    Renvoie la racine d'un arbre.

    Entrée : 
        arbre : l'arbre

    Exemples :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> ab = creer_arbre_binaire( s4 )
        >>> racine( ab ) is s4
        True
    """
    return arbre[0]

def etiquette( arbre, sommet ):
    """
    Renvoie l'etiquette d'un sommet.

    Entrée : 
        arbre : un arbre
        sommet : un sommet

    Exemples :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> ab = creer_arbre_binaire( s4 )
        >>> etiquette( ab, s8 ) == 8
        True
        >>> etiquette( ab, s9 ) == 9
        True
        >>> etiquette( ab, s4 ) == 4
        True
    """
    return sommet[1]


def parcours_infixe_sous_arbre( arbre, sommet, p ):
    """
    Rempli un liste 'p' avec des sommets du sous-arbre 'sommet' ordonné selon 
    le parcours infixe de l'arbre.

    Entrée :
        arbre : un arbre
        sommet : le sommet racine du sous-arbre
        p : une liste

    Exemples :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> l = []
        >>> parcours_infixe_sous_arbre( ab, racine(ab), l )
        >>> l == [s8, s4, s9, s2, s5, s11, s1, s12, s6, s3, s7]
        True
        >>> l = []
        >>> parcours_infixe_sous_arbre( ab, s3, l )
        >>> l == [s12, s6, s3, s7]
        True
    """
    fg = fils_gauche( arbre, sommet )
    fd = fils_droit( arbre, sommet )
    if not fg is None :
        parcours_infixe_sous_arbre( arbre, fg, p )
    p.append( sommet )
    if not fd is None:
        parcours_infixe_sous_arbre( arbre, fd, p )

def parcours_infixe( arbre, p ):
    """
    Rempli un liste 'p' avec des sommets de l'arbre 'arbre' ordonné selon 
    le parcours infixe de l'arbre.

    Entrée :
        arbre : un arbre
        p : une liste

    Exemples :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> l = []
        >>> parcours_infixe( ab, l )
        >>> l == [s8, s4, s9, s2, s5, s11, s1, s12, s6, s3, s7]
        True
    """
    parcours_infixe_sous_arbre( arbre, racine(arbre), p )

def parcours_prefixe_sous_arbre( arbre, sommet, p ):
    """
    Rempli un liste 'p' avec des sommets du sous-arbre 'sommet' ordonné selon 
    le parcours préfixe de l'arbre.

    Entrée :
        arbre : un arbre
        sommet : le sommet racine du sous-arbre
        p : une liste

    Exemples :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> l = []
        >>> parcours_prefixe_sous_arbre( ab, racine(ab), l )
        >>> l == [s1, s2, s4, s8, s9, s5, s11, s3, s6, s12, s7]
        True
        >>> l = []
        >>> parcours_prefixe_sous_arbre( ab, s3, l )
        >>> l == [s3, s6, s12, s7]
        True
    """
    fg = fils_gauche( arbre, sommet )
    fd = fils_droit( arbre, sommet )
    p.append( sommet )
    if not fg is None:
        parcours_prefixe_sous_arbre( arbre, fg, p )
    if not fd is None:
        parcours_prefixe_sous_arbre( arbre, fd, p )

def parcours_prefixe( arbre, p ):
    """
    Rempli un liste 'p' avec des sommets du sous-arbre 'sommet' ordonné selon 
    le parcours préfixe de l'arbre.

    Entrée :
        arbre : un arbre
        p : une liste

    Exemples :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> l = []
        >>> parcours_prefixe( ab, l )
        >>> l == [s1, s2, s4, s8, s9, s5, s11, s3, s6, s12, s7]
        True
    """
    parcours_prefixe_sous_arbre( arbre, racine(arbre), p )


def parcours_postfixe_sous_arbre( arbre, sommet, p ):
    """
    Rempli un liste 'p' avec des sommets du sous-arbre 'sommet' ordonné selon 
    le parcours postfixe de l'arbre.

    Entrée :
        arbre : un arbre
        sommet : le sommet racine du sous-arbre
        p : une liste

    Exemples :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> l = []
        >>> parcours_postfixe_sous_arbre( ab, racine(ab), l )
        >>> l == [s8, s9, s4, s11, s5, s2, s12, s6, s7, s3, s1]
        True
        >>> l = []
        >>> parcours_postfixe_sous_arbre( ab, s3, l )
        >>> l == [s12, s6, s7, s3]
        True
    """
    fg = fils_gauche( arbre, sommet )
    fd = fils_droit( arbre, sommet )
    if not fg is None:
        parcours_postfixe_sous_arbre( arbre, fg, p )
    if not fd is None:
        parcours_postfixe_sous_arbre( arbre, fd, p )
    p.append( sommet )

def parcours_postfixe( arbre, p ):
    """
    Rempli un liste 'p' avec des sommets de l'arbre ordonné selon 
    le parcours postfixe de l'arbre.

    Entrée :
        arbre : un arbre
        p : une liste

    Exemples :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> l = []
        >>> parcours_postfixe( ab, l )
        >>> l == [s8, s9, s4, s11, s5, s2, s12, s6, s7, s3, s1]
        True
    """
    parcours_postfixe_sous_arbre( arbre, racine(arbre), p )

def est_feuille( arbre, sommet ):
    """
    Renvoie vraie si le sommet est une feuille et non sinon.

    Entrée :
        arbre : un arbre
        sommet : un sommet 

    Exemples :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> est_feuille( ab, s8 )
        True
        >>> est_feuille( ab, s6 )
        False
        >>> est_feuille( ab, s5 )
        False
        >>> est_feuille( ab, s4 )
        False
    """
    return len( fils( arbre, sommet) ) == 0

def arbre_binaire_to_dot_rec( arbre, sommet, nb ):
    res = ""
    if est_feuille( arbre, sommet ):
        res += "\n    n" + str(nb) + ' [ label="' + str(etiquette(arbre, sommet))+ '" ];'
        return ( nb, res )
    fg = fils_gauche( arbre, sommet )
    fd = fils_droit( arbre, sommet )
    if not fg is None:
        ( nb_fg, dot_fg ) = arbre_binaire_to_dot_rec( arbre, fg, nb )
        res += dot_fg
        nb = nb_fg + 1
    else:
        res += "\n    n" + str(nb) + ' [ label="." ];'
        nb_fg = nb
        nb += 1
    if not fd is None:
        ( nb_fd, dot_fd ) = arbre_binaire_to_dot_rec( arbre, fd, nb )
        res += dot_fd
        nb = nb_fd + 1
    else:
        res += "\n    n" + str(nb) + ' [ label="." ];'
        nb_fd = nb
        nb += 1
    nb_racine = nb
    res += "\n    n" + str(nb_racine) + ' [ label="' + str(etiquette(arbre, sommet))+ '" ];'
    res += "\n    n" + str(nb_racine) + " -> " + "n" + str(nb_fg) + ";"
    res += "\n    n" + str(nb_racine) + " -> " + "n" + str(nb_fd) + ";"
    return ( nb_racine, res )


def arbre_binaire_to_dot( arbre ):
    """
    Renvoie l'arbre binaire en format dot.

    Entrée :
        arbre : un arbre

    Exemples :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> print( arbre_binaire_to_dot( ab ) )
        digraph G {
            n0 [ label="8" ];
            n1 [ label="9" ];
            n2 [ label="4" ];
            n2 -> n0;
            n2 -> n1;
            n3 [ label="." ];
            n4 [ label="11" ];
            n5 [ label="5" ];
            n5 -> n3;
            n5 -> n4;
            n6 [ label="2" ];
            n6 -> n2;
            n6 -> n5;
            n7 [ label="12" ];
            n8 [ label="." ];
            n9 [ label="6" ];
            n9 -> n7;
            n9 -> n8;
            n10 [ label="7" ];
            n11 [ label="3" ];
            n11 -> n9;
            n11 -> n10;
            n12 [ label="1" ];
            n12 -> n6;
            n12 -> n11;
        }
    """
    res = "digraph G {"
    res += arbre_binaire_to_dot_rec( arbre, racine(arbre), 0 )[1]
    res += "\n}"
    return res

def afficher_arbre_binaire( arbre ):
    """
    >>> s8 = creer_sommet( '8' )
    >>> s9 = creer_sommet( '9' )
    >>> s4 = creer_sommet( '4' )
    >>> s5 = creer_sommet( '5', None, s8 )
    >>> s6 = creer_sommet( '6' )
    >>> s7 = creer_sommet( '7', s9, None )
    >>> s2 = creer_sommet( '2', s4, s5 )
    >>> s3 = creer_sommet( '3', s6, s7 )
    >>> s1 = creer_sommet( '1', s2, s3 )
    >>> a = creer_arbre_binaire( s1 )
    >>> #afficher_arbre_binaire( a )
    """
    f = tempfile.NamedTemporaryFile( delete=False )
    f.write( bytes( arbre_binaire_to_dot( arbre ), 'UTF-8' ) )
    f.close()
    os.system( 'dotty ' + str(f.name) )
    os.remove( f.name )


def est_parfait_rec( arbre, sommet ):
    """
        Cette fonction renvoie un triplet (h, p, s) où h est la hauteur de 
        l'arbre, p est vrai si l'arbre est parfait, et s est vrai si le 
        dernier niveau est saturé.
    """
    if sommet is None:
        return (-1, True, True)
    (h_fg, p_fg, s_fg) = est_parfait_rec( arbre, fils_gauche( arbre, sommet ) )
    (h_fd, p_fd, s_fd) = est_parfait_rec( arbre, fils_droit( arbre, sommet ) )
    p = False
    s = False
    if h_fd == h_fg :
        if p_fg and s_fg and p_fd :
            p = True
            if s_fd :
                s = True
    elif h_fg == h_fd + 1:
        if p_fg and p_fd and s_fd:
            p = True
    return ( max( h_fd, h_fg ) + 1, p, s )

def est_parfait( arbre ):
    """
    >>> s8 = creer_sommet( '8' )
    >>> s9 = creer_sommet( '9' )
    >>> s4 = creer_sommet( '4' )
    >>> s5 = creer_sommet( '5', None, s8 )
    >>> s6 = creer_sommet( '6' )
    >>> s7 = creer_sommet( '7', s9, None )
    >>> s2 = creer_sommet( '2', s4, s5 )
    >>> s3 = creer_sommet( '3', s6, s7 )
    >>> s1 = creer_sommet( '1', s2, s3 )
    >>> a = creer_arbre_binaire( s1 )
    >>> est_parfait( a )
    False
    >>> a = creer_arbre_binaire( s3 )
    >>> est_parfait( a )
    False

    >>> s8 = creer_sommet( '8' )
    >>> s9 = creer_sommet( '9' )
    >>> s4 = creer_sommet( '4', s8, s9 )
    >>> s5 = creer_sommet( '5' )
    >>> s6 = creer_sommet( '6' )
    >>> s7 = creer_sommet( '7' )
    >>> s2 = creer_sommet( '2', s4, s5 )
    >>> s3 = creer_sommet( '3', s6, s7 )
    >>> s1 = creer_sommet( '1', s2, s3 )
    >>> a = creer_arbre_binaire( s1 )
    >>> est_parfait( a )
    True
    >>> a = creer_arbre_binaire( s2 )
    >>> est_parfait( a )
    True
    >>> a = creer_arbre_binaire( s4 )
    >>> est_parfait( a )
    True
    """
    return est_parfait_rec( arbre, racine(arbre) )[1]

def est_complet( arbre ):
    """
    >>> s8 = creer_sommet( '8' )
    >>> s9 = creer_sommet( '9' )
    >>> s4 = creer_sommet( '4' )
    >>> s5 = creer_sommet( '5', None, s8 )
    >>> s6 = creer_sommet( '6' )
    >>> s7 = creer_sommet( '7', s9, None )
    >>> s2 = creer_sommet( '2', s4, s5 )
    >>> s3 = creer_sommet( '3', s6, s7 )
    >>> s1 = creer_sommet( '1', s2, s3 )
    >>> a = creer_arbre_binaire( s1 )
    >>> est_complet( a )
    False
    >>> a = creer_arbre_binaire( s3 )
    >>> est_complet( a )
    False

    >>> s8 = creer_sommet( '8' )
    >>> s9 = creer_sommet( '9' )
    >>> s4 = creer_sommet( '4', s8, s9 )
    >>> s5 = creer_sommet( '5' )
    >>> s6 = creer_sommet( '6' )
    >>> s7 = creer_sommet( '7' )
    >>> s2 = creer_sommet( '2', s4, s5 )
    >>> s3 = creer_sommet( '3', s6, s7 )
    >>> s1 = creer_sommet( '1', s2, s3 )
    >>> a = creer_arbre_binaire( s1 )
    >>> est_complet( a )
    False
    >>> a = creer_arbre_binaire( s2 )
    >>> est_complet( a )
    False
    >>> a = creer_arbre_binaire( s4 )
    >>> est_complet( a )
    True
    """
    return est_parfait_rec( arbre, racine(arbre) )[2]

def copier_arbre_rec( arbre, sommet, copie_etiquette= lambda x: x ):
    """
    Renvoie une copie du sous-arbre binaire passée en paramètre.

    Entrée :
        arbre : un arbre
        sommet : le sommet du sous-arbre binaire

    Exemple :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> ab1 = copier_arbre_rec( ab, racine(ab) )
        >>> str(ab1) == str(racine(ab))
        True
    """
    if arbre is None or sommet is None:
        return None
    copie_gauche = copier_arbre_rec(
        arbre, fils_gauche(arbre, sommet), copie_etiquette
    )
    copie_droite = copier_arbre_rec(
        arbre, fils_droit(arbre, sommet), copie_etiquette
    )
    return creer_sommet( 
        copie_etiquette( etiquette(arbre, sommet) ), copie_gauche, copie_droite
    )

def copier_arbre( arbre, copie_etiquette= lambda x: x ):
    """
    Renvoie une copie de l'arbre biniaire passée en paramètre.

    Entrée :
        arbre : un arbre

    Exemple :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> ab1 = copier_arbre( ab )
        >>> str(ab1) == str(ab)
        True
    """
    return creer_arbre_binaire( 
        copier_arbre_rec( arbre, racine(arbre), copie_etiquette )
    )

def arbre_binaires_de_taille_fixe( n ):
    """
    Renvoie la liste de tous les arbres binaires de taille n.

    Entrée :
        arbre : un arbre

    Exemple :
        >>> l = arbre_binaires_de_taille_fixe( 1 )
        >>> len( l )
        1

        >>> l = arbre_binaires_de_taille_fixe( 2 )
        >>> len( l )
        2

        >>> l = arbre_binaires_de_taille_fixe( 3 )
        >>> len( l )
        5

        >>> l = arbre_binaires_de_taille_fixe( 4 )
        >>> len( l )
        14

        >>> l = arbre_binaires_de_taille_fixe( 5 )
        >>> len( l )
        42

    """
    if( n==0 ):
        return []
    if( n==1 ):
        return [ creer_arbre_binaire( creer_sommet("V") ) ]
    res = []
    for n_g in range( n ):
        n_d = n-1-n_g
        arbres_d = arbre_binaires_de_taille_fixe( n_d )
        arbres_g = arbre_binaires_de_taille_fixe( n_g )
        if n_d == 0 :
            arbres_d.append( creer_arbre_binaire(None) )
        if n_g == 0 :
            arbres_g.append( creer_arbre_binaire(None) )
        for a_g in arbres_g:
            for a_d in arbres_d:
                res.append(
                    creer_arbre_binaire( 
                        creer_sommet( "V", 
                            racine( copier_arbre( a_g ) ), 
                            racine( copier_arbre( a_d ) ) 
                        ) 
                    )
                )
    return res

def est_arbre_binaire_de_recherche_rec( arbre, sommet ):
    etiquette_racine = etiquette( arbre, sommet )
    maximal = etiquette_racine
    minimal = etiquette_racine
    recherche = True
    fg = fils_gauche( arbre, sommet )
    if not fg is None:
        (mi_g,ma_g,re_g) = est_arbre_binaire_de_recherche_rec( arbre, fg )
        recherche = recherche and ( ma_g <= etiquette_racine  ) and re_g
        minimal = mi_g
    fd = fils_droit( arbre, sommet )
    if not fd is None:
        (mi_d,ma_d,re_d) = est_arbre_binaire_de_recherche_rec( arbre, fd )
        recherche = recherche and ( mi_d > etiquette_racine  ) and re_d
        maximal = ma_d
    return ( minimal, maximal, recherche )

def est_arbre_binaire_de_recherche( arbre ):
    """
    Renvoie vraie si l'arbre est un arbre binaire de recherche.

    Entrée :
        arbre : un arbre

    Exemple :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> est_arbre_binaire_de_recherche( ab )
        False
        >>> s8 = creer_sommet( 1 )
        >>> s9 = creer_sommet( 3 )
        >>> s11 = creer_sommet( 6 )
        >>> s12 = creer_sommet( 8 )
        >>> s4 = creer_sommet( 2, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 9, s12, None )
        >>> s7 = creer_sommet( 11 )
        >>> s3 = creer_sommet( 10, s6, s7 )
        >>> s2 = creer_sommet( 4, s4, s5 )
        >>> s1 = creer_sommet( 7, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> est_arbre_binaire_de_recherche( ab )
        True

    """
    return est_arbre_binaire_de_recherche_rec( arbre, racine(arbre) )[2]


def rechercher_element_sous_arbre( arbre, sommet, e ):
    """
    Renvoie vraie si l'lement 'e' est dans le sous-arbre de sommet 'sommet'.

    Entrée :
        arbre : un arbre binaire de recherche
        sommet : un sommet
        e : l'element a rechercher

    Exemple :

        >>> s8 = creer_sommet( 1 )
        >>> s9 = creer_sommet( 3 )
        >>> s11 = creer_sommet( 7 )
        >>> s12 = creer_sommet( 9 )
        >>> s4 = creer_sommet( 2, s8, s9 )
        >>> s5 = creer_sommet( 6, None, s11 )
        >>> s6 = creer_sommet( 10, s12, None )
        >>> s7 = creer_sommet( 12 )
        >>> s3 = creer_sommet( 11, s6, s7 )
        >>> s2 = creer_sommet( 4, s4, s5 )
        >>> s1 = creer_sommet( 8, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> rechercher_element_sous_arbre( ab, s2, 7  )
        True
        >>> rechercher_element_sous_arbre( ab, s2, 5  )
        False

    """
    r = etiquette( arbre, sommet )
    if r == e:
        return True
    if r >= e:
        prochain = fils_gauche( arbre, sommet )
    else:
        prochain = fils_droit( arbre, sommet )
    return (
        ( not prochain is None ) and 
        rechercher_element_sous_arbre( arbre, prochain, e )
    )

def rechercher_element( arbre, e ):
    """
    Renvoie vraie si l'élément 'e' est dans l'arbre.

    Entrée :
        arbre : un arbre binaire de recherche
        e : l'element a rechercher

    Exemple :

        >>> s8 = creer_sommet( 1 )
        >>> s9 = creer_sommet( 3 )
        >>> s11 = creer_sommet( 7 )
        >>> s12 = creer_sommet( 9 )
        >>> s4 = creer_sommet( 2, s8, s9 )
        >>> s5 = creer_sommet( 6, None, s11 )
        >>> s6 = creer_sommet( 10, s12, None )
        >>> s7 = creer_sommet( 12 )
        >>> s3 = creer_sommet( 11, s6, s7 )
        >>> s2 = creer_sommet( 4, s4, s5 )
        >>> s1 = creer_sommet( 8, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> rechercher_element( ab, 7  )
        True
        >>> rechercher_element( ab, 5  )
        False

    """
    return rechercher_element_sous_arbre( arbre, racine(arbre), e )

def inserer_element_dans_arbre_binaire_de_recherche( arbre, e ):
    """
    Insere l'lément 'e' dans l'arbre binaire e.

    Entrée :
        arbre : un arbre binaire de recherche
        e : l'element à inserer

    Exemple :

        >>> ab = creer_arbre_binaire( None )
        >>> inserer_element_dans_arbre_binaire_de_recherche( ab, 4 )
        >>> inserer_element_dans_arbre_binaire_de_recherche( ab, 7 )
        >>> inserer_element_dans_arbre_binaire_de_recherche( ab, 2 )
        >>> inserer_element_dans_arbre_binaire_de_recherche( ab, 1 )
        >>> inserer_element_dans_arbre_binaire_de_recherche( ab, 6 )
        >>> inserer_element_dans_arbre_binaire_de_recherche( ab, 8 )
        >>> est_arbre_binaire_de_recherche( ab )
        True
        >>> rechercher_element( ab, 1 )
        True
        >>> rechercher_element( ab, 2 )
        True
        >>> rechercher_element( ab, 4 )
        True
        >>> rechercher_element( ab, 6 )
        True
        >>> rechercher_element( ab, 7 )
        True
        >>> rechercher_element( ab, 8 )
        True
        >>> taille_arbre_binaire( ab )
        6
    """
    s = creer_sommet( e )
    suivant = racine( arbre )
    if racine(arbre) is None:
        arbre[0] = s
        return
    sommet = None
    while( not suivant is None ):
        sommet = suivant
        if e <= etiquette( arbre, suivant ):
            suivant = fils_gauche( arbre, suivant )
        else:
            suivant = fils_droit( arbre, suivant )
    if e <= etiquette( arbre, sommet ):
        sommet[2][ 0 ] = s
    else:
        sommet[2][ 1 ] = s
    s[0] = sommet

def sous_arbre_est_un_tas( arbre, sommet ):
    """
    Renvoie Vrai si le sous-arbre de somme 'sommet' un tas
    Dans un tas l'etiquette dupère est toujours plus petite que les étiquettes
    de ses fils.

    Entrée :
        arbre - un arbre
        sommet - un sommet

    Exemple :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> sous_arbre_est_un_tas( ab, s2 )
        True

        >>> s8 = creer_sommet( 1 )
        >>> s9 = creer_sommet( 3 )
        >>> s11 = creer_sommet( 6 )
        >>> s12 = creer_sommet( 8 )
        >>> s4 = creer_sommet( 2, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 9, s12, None )
        >>> s7 = creer_sommet( 11 )
        >>> s3 = creer_sommet( 10, s6, s7 )
        >>> s2 = creer_sommet( 4, s4, s5 )
        >>> s1 = creer_sommet( 7, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> sous_arbre_est_un_tas( ab, s2 )
        False
    """
    etiquette_racine = etiquette( arbre, sommet )
    fg = fils_droit( arbre, sommet )
    tas = True
    fg = fils_gauche( arbre, sommet )
    if not fg is None:
        tas = tas and (
            etiquette_racine <= etiquette(arbre,fg) 
        ) and sous_arbre_est_un_tas( arbre, fg )
    fd = fils_droit( arbre, sommet )
    if not fd is None:
        tas = tas and ( 
            etiquette_racine <= etiquette(arbre, fd)
        ) and sous_arbre_est_un_tas( arbre, fd )
    return tas

def est_un_tas( arbre ):
    """
    Renvoie Vrai si l'arbre est un tas
    Dans un tas l'etiquette dupère est toujours plus petite que les étiquettes
    de ses fils.

    Entrée :
        arbre - un arbre

    Exemple :

        >>> s8 = creer_sommet( 8 )
        >>> s9 = creer_sommet( 9 )
        >>> s11 = creer_sommet( 11 )
        >>> s12 = creer_sommet( 12 )
        >>> s4 = creer_sommet( 4, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 6, s12, None )
        >>> s7 = creer_sommet( 7 )
        >>> s3 = creer_sommet( 3, s6, s7 )
        >>> s2 = creer_sommet( 2, s4, s5 )
        >>> s1 = creer_sommet( 1, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> est_un_tas( ab )
        True

        >>> s8 = creer_sommet( 1 )
        >>> s9 = creer_sommet( 3 )
        >>> s11 = creer_sommet( 6 )
        >>> s12 = creer_sommet( 8 )
        >>> s4 = creer_sommet( 2, s8, s9 )
        >>> s5 = creer_sommet( 5, None, s11 )
        >>> s6 = creer_sommet( 9, s12, None )
        >>> s7 = creer_sommet( 11 )
        >>> s3 = creer_sommet( 10, s6, s7 )
        >>> s2 = creer_sommet( 4, s4, s5 )
        >>> s1 = creer_sommet( 7, s2, s3 )
        >>> ab = creer_arbre_binaire( s1 )
        >>> est_un_tas( ab )
        False
    """
    return sous_arbre_est_un_tas( arbre, racine(arbre) )



if __name__ == "__main__":
    import doctest
    doctest.testmod() 
